import os

import click
import pygments

from .core import Combine
from .checks import CheckRunner
from .dev import Watcher, Server


@click.group()
@click.pass_context
def cli(ctx):
    pass


@cli.command()
@click.option("--no-checks", is_flag=True, default=False)
@click.option("--env", default="production")
@click.pass_context
def build(ctx, no_checks, env):
    config_path = os.path.abspath("combine.yml")
    combine = Combine(config_path=config_path, env=env)

    click.secho("Installing dependencies", fg="cyan")
    combine.install()

    click.secho("Building site", fg="cyan")
    combine.build()

    if not no_checks:
        runner = CheckRunner(combine)
        messages = runner.run()
        if messages:
            click.secho("Checks failed.", fg="red")
            for msg in messages:
                click.secho(str(msg), fg=msg.color)
                click.echo()
            exit(1)
        else:
            click.secho("All checks passed!", fg="green")

    return combine


@cli.command()
@click.option("--port", type=int, default=8000)
@click.pass_context
def work(ctx, port):
    combine = ctx.invoke(build, no_checks=True, env="development")

    click.secho("Watching for file changes...", fg="green")

    server = Server(combine.output_path, port)
    watcher = Watcher(".", combine=combine)
    watcher.watch(server.serve)


@cli.group()
@click.pass_context
def utils(ctx):
    pass


@utils.command()
@click.option("--style", type=str, default="default")
@click.pass_context
def highlight_info(ctx, style):
    """Outputs the CSS which can be customized for highlighted code"""
    click.secho("The following styles are available to choose from:", fg="green")
    click.echo(list(pygments.styles.get_all_styles()))
    click.echo()
    click.secho(
        f'The following CSS for the "{style}" style can be customized:', fg="green"
    )
    click.echo(pygments.formatters.HtmlFormatter(style=style).get_style_defs())


# @cli.command()
# @click.pass_context
# def check(ctx):
#     combine = ctx.invoke(build, no_checks=True)  # should it build?
#     runner = CheckRunner(combine)
#     runner.run()
#     if runner.succeeded:
#         click.secho('All checks passed!', fg='green')
#     else:
#         click.secho('Checks failed.', fg='red')
#         for check in runner.failed_checks:
#             click.echo(check)
#         exit(1)


if __name__ == "__main__":
    cli()


# Want to make it as easy and friendly as possible, hard to screw up
# pretty print output html? or don't want to screw with js, etc.
# validate html
# check "output" in .gitignore (if git) -- use warning codes FW101
# asset pipeline (asset fingerprinting - hash)

# rel canonical?

# build in pre-commit hook helper -- you don't have a pre-commit hook, do you want one?
# then it runs check command before commit (not on every run if costly)

# no links to same page (empty links too)



from jinja2 import nodes
from jinja2.ext import Extension

from pygments import highlight
from pygments.lexers import get_lexer_by_name, guess_lexer
from pygments.formatters import HtmlFormatter


class CodeHighlightExtension(Extension):
    tags = set(["code"])

    def __init__(self, environment):
        super().__init__(environment)

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        args = [parser.parse_expression()]
        body = parser.parse_statements(["name:endcode"], drop_needle=True)
        return nodes.CallBlock(
            self.call_method("_code_support", args), [], [], body
        ).set_lineno(lineno)

    def _code_support(self, language, caller):
        """Helper callback."""
        code = caller()

        # remove the leading whitespace so the whole block can be indented more easily with flow of page
        lines = code.splitlines()

        first_nonempty_line_index = 0
        while not lines[first_nonempty_line_index]:
            first_nonempty_line_index += 1

        len_to_trim = len(lines[first_nonempty_line_index]) - len(
            lines[first_nonempty_line_index].lstrip()
        )
        lines = [x[len_to_trim:] for x in lines]
        code = "\n".join(lines)

        if language:
            lexer = get_lexer_by_name(language, stripall=True)
        else:
            lexer = guess_lexer(code)

        highlighted = highlight(code, lexer, HtmlFormatter())

        return highlighted



from .registry import registry


class CheckRunner:
    def __init__(self, combine):
        self.combine = combine
        self.registry = registry
        self.checks = [x(self.combine) for x in self.registry.registered_checks]

    def run(self):
        all_messages = []

        for check in self.checks:
            messages = check.run()
            if messages:
                all_messages = all_messages + messages

        return all_messages



from .core import Check


class CheckRegistry:
    def __init__(self):
        self.registered_checks = set()

    def register(self, check):
        assert issubclass(check, Check), "Check must be an instance of Check"
        self.registered_checks.add(check)
        return check


registry = CheckRegistry()
register = registry.register



from jinja2 import nodes
from jinja2.ext import Extension

import markdown
from markdown.extensions.codehilite import CodeHiliteExtension


class MarkdownExtension(Extension):
    tags = set(["markdown"])

    def __init__(self, environment):
        super().__init__(environment)

    def parse(self, parser):
        lineno = next(parser.stream).lineno
        body = parser.parse_statements(["name:endmarkdown"], drop_needle=True)
        return nodes.CallBlock(self.call_method("_support"), [], [], body).set_lineno(
            lineno
        )

    def _support(self, caller):
        """Helper callback."""
        markdown_content = caller()
        html_content = markdown.markdown(
            markdown_content,
            extensions=[
                "markdown.extensions.fenced_code",
                CodeHiliteExtension(css_class="highlight"),
                "markdown.extensions.tables",
            ],
        )
        return html_content
